/*
 * Copyright (C) 2018 Red Hat, Inc. (www.redhat.com)
 * SPDX-FileCopyrightText: (C) 2024 Siemens AG
 *
 * This library is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#include "evolution-ews-config.h"

#include <glib/gi18n-lib.h>
#include <libedataserver/libedataserver.h>

#include "server/camel-ews-settings.h"
#include "e-ews-debug.h"
#include "e-ms-oapxbc-util.h"

#include "e-oauth2-service-office365.h"

/* https://portal.azure.com/
   https://docs.microsoft.com/en-us/azure/active-directory/develop/active-directory-developers-guide
   https://tsmatz.wordpress.com/2016/10/07/application-permission-with-v2-endpoint-and-microsoft-graph/
*/

struct _EOAuth2ServiceOffice365Private
{
	GMutex string_cache_lock;
	GHashTable *string_cache;
};

/* Forward Declarations */
static void e_oauth2_service_office365_oauth2_service_init (EOAuth2ServiceInterface *iface);

G_DEFINE_DYNAMIC_TYPE_EXTENDED (EOAuth2ServiceOffice365, e_oauth2_service_office365, E_TYPE_OAUTH2_SERVICE_BASE, 0,
	G_IMPLEMENT_INTERFACE_DYNAMIC (E_TYPE_OAUTH2_SERVICE, e_oauth2_service_office365_oauth2_service_init))

static const gchar *
eos_office365_cache_string (EOAuth2ServiceOffice365 *oauth2_office365,
			    gchar *str) /* takes ownership of the 'str' */
{
	const gchar *cached_str;

	g_return_val_if_fail (E_IS_OAUTH2_SERVICE_OFFICE365 (oauth2_office365), NULL);

	if (!str)
		return NULL;

	if (!*str) {
		g_free (str);
		return "";
	}

	g_mutex_lock (&oauth2_office365->priv->string_cache_lock);

	cached_str = g_hash_table_lookup (oauth2_office365->priv->string_cache, str);
	if (cached_str) {
		g_free (str);
	} else {
		g_hash_table_insert (oauth2_office365->priv->string_cache, str, str);
		cached_str = str;
	}

	g_mutex_unlock (&oauth2_office365->priv->string_cache_lock);

	return cached_str;
}

static const gchar *
eos_office365_get_endpoint_host (EOAuth2ServiceOffice365 *oauth2_office365,
				 CamelEwsSettings *ews_settings)
{
	if (ews_settings && camel_ews_settings_get_override_oauth2 (ews_settings)) {
		gchar *endpoint_host;

		endpoint_host = camel_ews_settings_dup_oauth2_endpoint_host (ews_settings);

		if (endpoint_host && *endpoint_host)
			return eos_office365_cache_string (oauth2_office365, endpoint_host);

		g_free (endpoint_host);
	}

	return OFFICE365_ENDPOINT_HOST;
}

static CamelEwsSettings *
eos_office365_get_camel_settings (ESource *source)
{
	ESourceCamel *extension;

	if (!source)
		return NULL;

	g_return_val_if_fail (E_IS_SOURCE (source), NULL);

	extension = e_source_get_extension (source, e_source_camel_get_extension_name ("ews"));

	return CAMEL_EWS_SETTINGS (e_source_camel_get_settings (extension));
}

static gboolean
eos_office365_guess_can_process (EOAuth2Service *service,
				 const gchar *protocol,
				 const gchar *hostname)
{
	return e_oauth2_services_is_supported () &&
		protocol && g_ascii_strcasecmp (protocol, "ews") == 0;
}

static const gchar *
eos_office365_get_name (EOAuth2Service *service)
{
	return "Office365";
}

static const gchar *
eos_office365_get_display_name (EOAuth2Service *service)
{
	/* Translators: This is a user-visible string, display name of an OAuth2 service. */
	return C_("OAuth2Service", "Office365");
}

static const gchar *
eos_office365_get_client_id (EOAuth2Service *service,
			     ESource *source)
{
	EOAuth2ServiceOffice365 *oauth2_office365 = E_OAUTH2_SERVICE_OFFICE365 (service);
	CamelEwsSettings *ews_settings;

	ews_settings = eos_office365_get_camel_settings (source);
	if (ews_settings && camel_ews_settings_get_override_oauth2 (ews_settings)) {
		gchar *client_id = camel_ews_settings_dup_oauth2_client_id (ews_settings);

		if (client_id && !*client_id) {
			g_free (client_id);
			client_id = NULL;
		}

		if (client_id)
			return eos_office365_cache_string (oauth2_office365, client_id);
	}

	return OFFICE365_CLIENT_ID;
}

static const gchar *
eos_office365_get_client_secret (EOAuth2Service *service,
				 ESource *source)
{
	return NULL;
}

static const gchar *
eos_office365_get_authentication_uri (EOAuth2Service *service,
				      ESource *source)
{
	EOAuth2ServiceOffice365 *oauth2_office365 = E_OAUTH2_SERVICE_OFFICE365 (service);
	CamelEwsSettings *ews_settings;

	ews_settings = eos_office365_get_camel_settings (source);
	if (ews_settings && camel_ews_settings_get_override_oauth2 (ews_settings)) {
		gchar *tenant;
		const gchar *res;

		tenant = camel_ews_settings_dup_oauth2_tenant (ews_settings);
		if (tenant && !*tenant) {
			g_free (tenant);
			tenant = NULL;
		}

		res = eos_office365_cache_string (oauth2_office365,
			g_strdup_printf ("https://%s/%s/oauth2/authorize",
				eos_office365_get_endpoint_host (oauth2_office365, ews_settings),
				tenant ? tenant : OFFICE365_TENANT));

		g_free (tenant);

		return res;
	}

	return eos_office365_cache_string (oauth2_office365,
		g_strdup_printf ("https://%s/%s/oauth2/authorize",
			eos_office365_get_endpoint_host (oauth2_office365, ews_settings),
			OFFICE365_TENANT));
}

static const gchar *
eos_office365_get_refresh_uri (EOAuth2Service *service,
			       ESource *source)
{
	EOAuth2ServiceOffice365 *oauth2_office365 = E_OAUTH2_SERVICE_OFFICE365 (service);
	CamelEwsSettings *ews_settings;

	ews_settings = eos_office365_get_camel_settings (source);
	if (ews_settings && camel_ews_settings_get_override_oauth2 (ews_settings)) {
		gchar *tenant;
		const gchar *res;

		tenant = camel_ews_settings_dup_oauth2_tenant (ews_settings);
		if (tenant && !*tenant) {
			g_free (tenant);
			tenant = NULL;
		}

		res = eos_office365_cache_string (oauth2_office365,
			g_strdup_printf ("https://%s/%s/oauth2/token",
				eos_office365_get_endpoint_host (oauth2_office365, ews_settings),
				tenant ? tenant : OFFICE365_TENANT));

		g_free (tenant);

		return res;
	}

	return eos_office365_cache_string (oauth2_office365,
		g_strdup_printf ("https://%s/%s/oauth2/token",
			eos_office365_get_endpoint_host (oauth2_office365, ews_settings),
			OFFICE365_TENANT));
}

static const gchar *
eos_office365_get_redirect_uri (EOAuth2Service *service,
				ESource *source)
{
	EOAuth2ServiceOffice365 *oauth2_office365 = E_OAUTH2_SERVICE_OFFICE365 (service);
	CamelEwsSettings *ews_settings;
	const gchar *res;

	ews_settings = eos_office365_get_camel_settings (source);
	if (ews_settings && camel_ews_settings_get_override_oauth2 (ews_settings)) {
		gchar *redirect_uri;

		redirect_uri = camel_ews_settings_dup_oauth2_redirect_uri (ews_settings);

		if (redirect_uri && !*redirect_uri) {
			g_free (redirect_uri);
			redirect_uri = NULL;
		}

		if (redirect_uri)
			return eos_office365_cache_string (oauth2_office365, redirect_uri);

		if (e_ews_util_strcmp0 (camel_ews_settings_get_oauth2_endpoint_host (ews_settings), NULL) != 0) {
			return eos_office365_cache_string (oauth2_office365,
				g_strdup_printf ("https://%s/common/oauth2/nativeclient",
					eos_office365_get_endpoint_host (oauth2_office365, ews_settings)));
		}
	}

	res = OFFICE365_REDIRECT_URI;
	if (res && *res)
		return res;

	return eos_office365_cache_string (oauth2_office365,
		g_strdup_printf ("https://%s/common/oauth2/nativeclient",
			eos_office365_get_endpoint_host (oauth2_office365, ews_settings)));
}

static const gchar *
eos_office365_get_resource_uri (EOAuth2Service *service,
				ESource *source)
{
	EOAuth2ServiceOffice365 *oauth2_office365 = E_OAUTH2_SERVICE_OFFICE365 (service);
	CamelEwsSettings *ews_settings;

	ews_settings = eos_office365_get_camel_settings (source);
	if (ews_settings && camel_ews_settings_get_override_oauth2 (ews_settings)) {
		gchar *resource_uri;

		resource_uri = camel_ews_settings_dup_oauth2_resource_uri (ews_settings);

		if (resource_uri && !*resource_uri) {
			g_free (resource_uri);
			resource_uri = NULL;
		}

		if (resource_uri)
			return eos_office365_cache_string (oauth2_office365, resource_uri);
	}

	if (ews_settings) {
		gchar *host_url;

		host_url = camel_ews_settings_dup_hosturl (ews_settings);

		if (host_url && *host_url) {
			gchar *ptr;

			ptr = strstr (host_url, "://");
			ptr = ptr ? strchr (ptr + 3, '/') : NULL;

			if (ptr) {
				*ptr = '\0';

				return eos_office365_cache_string (oauth2_office365, host_url);
			}
		}

		g_free (host_url);
	}

	return OFFICE365_FALLBACK_RESOURCE_URI;
}

static void
eos_office365_prepare_authentication_uri_query (EOAuth2Service *service,
						ESource *source,
						GHashTable *uri_query)
{
	g_return_if_fail (uri_query != NULL);

	e_oauth2_service_util_set_to_form (uri_query, "response_mode", "query");
	e_oauth2_service_util_set_to_form (uri_query, "prompt", "login");
	e_oauth2_service_util_set_to_form (uri_query, "resource", eos_office365_get_resource_uri (service, source));
}

static gboolean
eos_office365_extract_authorization_code (EOAuth2Service *service,
					  ESource *source,
					  const gchar *page_title,
					  const gchar *page_uri,
					  const gchar *page_content,
					  gchar **out_authorization_code)
{
	SoupURI *suri;
	gboolean known = FALSE;

	g_return_val_if_fail (out_authorization_code != NULL, FALSE);

	*out_authorization_code = NULL;

	if (!page_uri || !*page_uri)
		return FALSE;

	suri = soup_uri_new (page_uri);
	if (!suri)
		return FALSE;

	if (suri->query) {
		GHashTable *uri_query = soup_form_decode (suri->query);

		if (uri_query) {
			const gchar *code;

			code = g_hash_table_lookup (uri_query, "code");

			if (code && *code) {
				*out_authorization_code = g_strdup (code);
				known = TRUE;
			} else if (g_hash_table_lookup (uri_query, "error")) {
				known = TRUE;
				if (g_strcmp0 (g_hash_table_lookup (uri_query, "error"), "access_denied") != 0) {
					const gchar *description;

					description = g_hash_table_lookup (uri_query, "error_description");
					if (description) {
						g_warning ("%s: error:%s description:%s", G_STRFUNC,
							(const gchar *) g_hash_table_lookup (uri_query, "error"),
							description);
					}
				}
			}

			g_hash_table_unref (uri_query);
		}
	}

	soup_uri_free (suri);

	return known;
}

static void
eos_office365_prepare_refresh_token_form (EOAuth2Service *service,
					  ESource *source,
					  const gchar *refresh_token,
					  GHashTable *form)
{
	g_return_if_fail (form != NULL);

	e_oauth2_service_util_set_to_form (form, "resource", eos_office365_get_resource_uri (service, source));
	e_oauth2_service_util_set_to_form (form, "redirect_uri", e_oauth2_service_get_redirect_uri (service, source));
}

static void
eos_office365_finalize (GObject *object)
{
	EOAuth2ServiceOffice365 *oauth2_office365 = E_OAUTH2_SERVICE_OFFICE365 (object);

	g_mutex_lock (&oauth2_office365->priv->string_cache_lock);
	g_hash_table_destroy (oauth2_office365->priv->string_cache);
	g_mutex_unlock (&oauth2_office365->priv->string_cache_lock);
	g_mutex_clear (&oauth2_office365->priv->string_cache_lock);

	/* Chain up to parent's method. */
	G_OBJECT_CLASS (e_oauth2_service_office365_parent_class)->finalize (object);
}

static SoupCookie *
eos_office365_get_prt_sso_cookie_sync (EOAuth2Service *service,
				       ESource *source,
				       GCancellable *cancellable)
{
	SoupCookie *prtsso_cookie = NULL;
	const gchar *sso_uri = eos_office365_get_authentication_uri (service, source);
	const gchar *client_id = eos_office365_get_client_id (service, source);
	const gchar *redirect_uri = eos_office365_get_redirect_uri (service, source);
	CamelEwsSettings *ews_settings = NULL;
	gchar *auth_user = NULL;
	GError *local_error = NULL;

	ews_settings = eos_office365_get_camel_settings (source);
	auth_user = camel_network_settings_dup_user (CAMEL_NETWORK_SETTINGS (ews_settings));

	prtsso_cookie = e_ms_oapxbc_util_get_prt_sso_cookie_sync (service, source, client_id, sso_uri,
								  redirect_uri, auth_user, cancellable, &local_error);
	g_free (auth_user);
	if (!prtsso_cookie) {
		/* not available in this version */
		/* e_ews_debug_print (_("Failed to acquire PRT SSO cookie: %s"), local_error ? local_error->message : "Unknown error"); */
		g_clear_error (&local_error);
	}

	return prtsso_cookie;
}

static GSList *
eos_office365_dup_credentials_prompter_cookies_sync (EOAuth2Service *service,
						     ESource *source,
						     GCancellable *cancellable)
{
	SoupCookie *prtsso_cookie;
	GSList *cookies = NULL;
	prtsso_cookie = eos_office365_get_prt_sso_cookie_sync (service, source, cancellable);
	if (!prtsso_cookie) {
		return NULL;
	}
	cookies = g_slist_append (cookies, prtsso_cookie);
	return cookies;
}

static void
eos_office365_prepare_refresh_token_message (EOAuth2Service *service,
					     ESource *source,
					     SoupMessage *message)
{
	SoupMessageHeaders *request_headers;
	SoupCookie *cookie;

	cookie = eos_office365_get_prt_sso_cookie_sync (service, source, NULL);
	if (!cookie) {
		return;
	}
	request_headers = message->request_headers;
	soup_message_headers_append (request_headers,
				     soup_cookie_get_name (cookie),
				     soup_cookie_get_value (cookie));
	soup_cookie_free (cookie);
}

static void
e_oauth2_service_office365_oauth2_service_init (EOAuth2ServiceInterface *iface)
{
	iface->guess_can_process = eos_office365_guess_can_process;
	iface->get_name = eos_office365_get_name;
	iface->get_display_name = eos_office365_get_display_name;
	iface->get_client_id = eos_office365_get_client_id;
	iface->get_client_secret = eos_office365_get_client_secret;
	iface->get_authentication_uri = eos_office365_get_authentication_uri;
	iface->get_refresh_uri = eos_office365_get_refresh_uri;
	iface->get_redirect_uri = eos_office365_get_redirect_uri;
	iface->prepare_authentication_uri_query = eos_office365_prepare_authentication_uri_query;
	iface->extract_authorization_code = eos_office365_extract_authorization_code;
	iface->prepare_refresh_token_form = eos_office365_prepare_refresh_token_form;
	iface->prepare_refresh_token_message = eos_office365_prepare_refresh_token_message;
	iface->dup_credentials_prompter_cookies_sync = eos_office365_dup_credentials_prompter_cookies_sync;
}

static void
e_oauth2_service_office365_class_init (EOAuth2ServiceOffice365Class *klass)
{
	GObjectClass *object_class;

	g_type_class_add_private (klass, sizeof (EOAuth2ServiceOffice365Private));

	object_class = G_OBJECT_CLASS (klass);
	object_class->finalize = eos_office365_finalize;
}

static void
e_oauth2_service_office365_class_finalize (EOAuth2ServiceOffice365Class *klass)
{
}

static void
e_oauth2_service_office365_init (EOAuth2ServiceOffice365 *oauth2_office365)
{
	oauth2_office365->priv = G_TYPE_INSTANCE_GET_PRIVATE (oauth2_office365, E_TYPE_OAUTH2_SERVICE_OFFICE365, EOAuth2ServiceOffice365Private);

	g_mutex_init (&oauth2_office365->priv->string_cache_lock);
	oauth2_office365->priv->string_cache = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
}

void
e_oauth2_service_office365_type_register (GTypeModule *type_module)
{
	e_oauth2_service_office365_register_type (type_module);
}

gint
e_ews_util_strcmp0 (const gchar *str1,
		    const gchar *str2)
{
	if (str1 && !*str1)
		str1 = NULL;

	if (str2 && !*str2)
		str2 = NULL;

	return g_strcmp0 (str1, str2);
}
